import asyncio
import functools

import rx
import rx.operators as ops
from rx.core.notification import OnCompleted, OnError, OnNext
from rx.disposable import Disposable
from rx.scheduler.eventloop import AsyncIOScheduler, AsyncIOThreadSafeScheduler


class class_or_instance_method(classmethod):
    def __get__(self, instance, type_):
        descr_get = super().__get__ if instance is None else self.__func__.__get__
        return descr_get(instance, type_)


def async_iterable_to_observable(iter, loop):
    def on_subscribe(observer, scheduler):
        async def _aio_sub():
            try:
                async for i in iter:
                    observer.on_next(i)
                loop.call_soon(observer.on_completed)
            except Exception as e:
                loop.call_soon(functools.partial(observer.on_error, e))

        task = asyncio.ensure_future(_aio_sub(), loop=loop)
        return Disposable(lambda: task.cancel())

    return rx.create(on_subscribe)


async def observable_to_async_iterable(obs, loop):
    queue = asyncio.Queue()

    def on_next(i):
        loop.call_soon_threadsafe(queue.put_nowait, i)

    disposable = obs.pipe(ops.materialize()).subscribe(
        on_next=on_next, scheduler=AsyncIOScheduler(loop=loop)
    )

    while True:
        i = await queue.get()
        if isinstance(i, OnNext):
            yield i.value
            queue.task_done()
        elif isinstance(i, OnError):
            disposable.dispose()
            raise (Exception(i.value))
        else:
            disposable.dispose()
            break



import rx
from rx import operators as ops

from rxpipes import Pipeline


class ScheduleEach(Pipeline):
    """
    Maps each element using operation on scheduler. Order is not preserved.
    """

    def setup(self, operation, scheduler):
        """"""
        self.scheduler = scheduler
        self.operation = operation

    def transform(self):
        return ops.flat_map(
            lambda x: rx.of(x).pipe(
                ops.map(self.operation), ops.subscribe_on(self.scheduler)
            )
        )


class Parallel(ScheduleEach):
    """
    Maps each element using operation on scheduler. Order is preserved.
    """

    def setup(self, operation, scheduler):
        """"""
        self.counter = 0

        def f(x):
            self.counter += 1
            return (self.counter, operation(x))

        super().setup(f, scheduler)

    def transform(self):
        return rx.pipe(
            super().transform(),
            ops.to_iterable(),
            ops.map(lambda xs: [y[1] for y in sorted(xs, key=lambda x: x[0])]),
        )



from distutils.core import setup

from setuptools import find_packages

exec(open("rxpipes/version.py").read())
setup(
    name="rxpipes",
    version=__version__,
    author="shirecoding",
    author_email="shirecoding@gmail.com",
    install_requires=["rx", "toolz"],
    extras_require={
        "test": [
            "pytest",
            "pytest-cov",
            "pytest-html",
            "pytest-metadata",
            "numpy",
            "opencv-python",
        ],
        "docs": ["mkdocs", "mkdocstrings", "mkdocs-material"],
    },
    url="https://github.com/shirecoding/RxPipes",
    download_url=f"https://github.com/shirecoding/ModelMaker/archive/{__version__}.tar.gz",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    zip_safe=False,
    include_package_data=True,
    packages=find_packages() + [],
    package_data={},
)



import logging
from typing import Any, Callable, Iterable, Optional, Type

import rx
from rx import Observable, operators
from rx.scheduler.eventloop import AsyncIOScheduler
from rx.subject import Subject

from .utils import class_or_instance_method

log = logging.getLogger(__name__)


class Pipeline:
    def __init__(self, *args: Optional[Any], **kwargs: Optional[Any]):
        """Pipeline

        Args and kwargs will be forwarded to user defined setup

        Args:
            args: args passed to user defined setup
            kwargs: kwargs passed to user defined setup
        """

        # call user setup
        self.setup(*args, **kwargs)

    ##############################################################################
    ## OVERRIDE WHEN SUBCLASSING
    ##############################################################################

    def setup(self, *args: Optional[Any], **kwargs: Optional[Any]):
        """
        Override this setup function to implement custom functionality when subclassing Pipeline

        Args:
            args: user defined args
            kwargs: user defined kwards
        """

    def transform(self) -> Callable[[rx.typing.Observable], rx.typing.Observable]:
        """
        Override this transform function to implement custom functionality when subclassing Pipeline

        Returns:
            a function mapping an observable to another (default = lambda x: x)
        """
        return lambda x: x

    ##############################################################################
    ## INTERNALS
    ##############################################################################

    def _operation(self):
        return operators.pipe(self.transform())

    ##############################################################################
    ## USAGE
    ##############################################################################

    @class_or_instance_method
    def pipe(self, *pipelines: "Pipeline") -> "Pipeline":
        """

        Can be used as a class or instance method to create a new pipeline chain

        Args:
            pipelines: variable number of pipelines

        Returns:
            newly composed Pipeline instance
        """

        # called as instance method
        if not isinstance(self, type):
            parent = self
            return type(
                "Pipeline",
                (Pipeline,),
                {
                    "transform": lambda self: rx.pipe(
                        parent.transform(), *[p.transform() for p in pipelines]
                    )
                },
            )()

        # called as class method
        else:
            return type(
                "Pipeline",
                (Pipeline,),
                {
                    "transform": lambda self: rx.pipe(
                        *[p.transform() for p in pipelines]
                    )
                },
            )()

    def to_observable(self, *args):
        return rx.merge(
            *[
                x
                if isinstance(x, rx.core.typing.Observable)
                else rx.from_(x)
                if isinstance(x, Iterable)
                else rx.of(x)
                for x in args
            ]
        ).pipe(self._operation())

    def to_blocking(self, *args):
        return self.to_observable(*args).pipe(operators.to_list()).run()

    def __call__(self, *args):
        """
        Alias for to_blocking
        """

        return self.to_blocking(*args)


##############################################################################
## Inject operators
##############################################################################


def pipeline_from_operator_jit(
    op: Callable[[rx.typing.Observable], rx.typing.Observable]
):
    """
    Returns a function that when called returns a Pipeline instance which
    pipes the parent transform to the injected operator

    Args:
        op: rx operator

    Returns:
        function returning a just-in-time created Pipeline instance
    """

    @class_or_instance_method
    def _f(parent, *args, **kwargs):
        if not isinstance(parent, type):
            return type(
                "Pipeline",
                (Pipeline,),
                {
                    "transform": lambda _: rx.pipe(
                        parent.transform(), op(*args, **kwargs)
                    )
                },
            )()
        else:
            return type(
                "Pipeline",
                (Pipeline,),
                {"transform": lambda _: op(*args, **kwargs)},
            )()

    return _f


# inject operators into Pipeline class
for op in [
    x for x in dir(operators) if x[0] != "_" and x[0].islower() and x not in ["pipe"]
]:
    setattr(Pipeline, op, pipeline_from_operator_jit(getattr(operators, op)))
